﻿
--Create the addon using Ace3
BRAnkhTimer = LibStub("AceAddon-3.0"):NewAddon("Broker_AnkhTimer", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")

--Create the data broker
BRAnkhTimer.dataObject = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("AnkhTimer", {
	type = "data source",
	text = "Ankhs",
    icon = "Interface\\AddOns\\Broker_AnkhTimer\\icon.tga",
    label = "AnkhTimer"
})


--Localization
local L = LibStub("AceLocale-3.0"):GetLocale("Broker_AnkhTimer", true)

--Local variables
local ankhsInBags = 0
local reincarnationCooldown = 3600
local glyphPresent = false
local timerHandle = nil

--RGB color codes
local COLOR_HEX_RED       = "ff0000"
local COLOR_HEX_ORANGE    = "ff7f00"
local COLOR_HEX_YELLOW    = "ffff00"
local COLOR_HEX_GREEN     = "00ff00"
local COLOR_HEX_WHITE     = "ffffff"
local COLOR_HEX_COPPER    = "eda55f"
local COLOR_HEX_SILVER    = "c7c7cf"
local COLOR_HEX_GOLD      = "ffd700"
local COLOR_HEX_PURPLE    = "9980CC"
local COLOR_HEX_BLUE	  = "0000ff"
local COLOR_HEX_BLACK	  = "000000"


--###############################################################
--AceAddon call back fuctions
--###############################################################


function BRAnkhTimer:OnInitialize()
	
	--Define configuration options
	local options = {
    	type ="group",	
		name = "AnkhTimer",
    	handler = BRAnkhTimer,
    	childGroups = "tree",
    	args = {
    		header1 = {
				type = "description",
				name = "v0.9.1",
				order = 1,
				width = "full",
				cmdHidden = true
			},
    		showCount = {
            	type = "toggle",
            	name = L["show_count"],
            	order = 2,
				width = "full",
            	desc = L["show_count_desc"],
           	 	set = function (info, value) 
           	 			self.db.profile.showCount = value 
           	 			self:RefreshText()
           	 		  end,
            	get = function (info) return self.db.profile.showCount end
        	},
        	showCooldown = {
            	type = "toggle",
            	name = L["show_cooldown"],
            	order = 3,
				width = "full",
            	desc = L["show_cooldown_desc"],
           	 	set = function (info, value) 
           	 			self.db.profile.showCooldown = value 
           	 			self:RefreshText()
           	 		  end,
            	get = function (info) return self.db.profile.showCooldown end
        	},
        	autoBuyAnkhs = {
            	type = "toggle",
            	name = L["buy_ankhs"],
            	order = 4,
				width = "full",
            	desc = L["buy_ankhs_desc"],
           	 	set = function (info, value) self.db.profile.autoBuyAnkhs = value end,
            	get = function (info) return self.db.profile.autoBuyAnkhs end
        	},
    		maxAnkhs = {
            	type = "input",
            	order = 5,
            	name = L["max_ankhs"],
            	desc = L["max_ankhs_desc"],
           	 	set = "SetMaxAnkhs",
            	get = "GetMaxAnkhs"
        	},
        	header2 = {
				type = "description",
				name = L["glyph_of_renewed_life_desc"],
				order = 6,
				width = "full",
				cmdHidden = true
			},
        	testPrint = {
            	type = "execute",
            	name = "TestPrint",
            	order = 7,
            	hidden = true,
            	desc = "Print out the internal state of the addon",
           	 	func = "TestPrint"
        	},
        	debugPrints = {
            	type = "toggle",
            	name = "Debug Prints",
            	order = 8,
            	hidden = true,
            	desc = "Enable debug prints",
           	 	set = function (info, value) self.db.profile.debugPrints = value end,
            	get = function (info) return self.db.profile.debugPrints end
            }
    	}
	}
	
	
	--Register options and set up configuration options
	local AceCfg = LibStub("AceConfig-3.0")
	local AceCfgReg = LibStub("AceConfigRegistry-3.0")
	local AceDlg = LibStub("AceConfigDialog-3.0")
	
	--Register options and slash commands
	AceCfg:RegisterOptionsTable("Broker_AnkhTimer", options, {"Broker_AnkhTimer", "BRAT"})
	
	--Set up configuration using Blizzards addon tab
	--Check if "Broker" group exists, in not create one
	local brokerOptions = AceCfgReg:GetOptionsTable("Broker", "dialog", "LibDataBroker-1.1")
	if (not brokerOptions) then
		brokerOptions = {
			type = "group",
			name = "Broker",
			args = {
			}
		}
		AceCfg:RegisterOptionsTable("Broker", brokerOptions)
		AceDlg:AddToBlizOptions("Broker", "Broker")
	end
	--Add addon configuration options
	self.optionsFrame = AceDlg:AddToBlizOptions("Broker_AnkhTimer", "AnkhTimer", "Broker")
	
	--Default database values
	local dbDefaults = {
		profile = {
			debugPrints = false,
			showCount = true,
			showCooldown = true,
			autoBuyAnkhs = true,
			maxAnkhs = 10,
			reincarnationLastUsed = 0
		}
	}	
		
	--Set up the database
	self.db = LibStub("AceDB-3.0"):New("Broker_AnkhTimerDB", dbDefaults)
	
	--The printout needs to be here
	self:DebugPrint("OnInitialize")
	
end


function BRAnkhTimer:OnEnable()
	self:DebugPrint("OnEnable")
	
	--Start a timer and run part of the functionality after the UI if fully initialized
	--This delay is needed since its not possible to read any talent information at the OnEnable state (I didnt have this problem before)
	timerHandle = self:ScheduleTimer("DelayedOnEnable", 1)
	
	--Checking when spell and talent changes, see if the Reincarnat cooldown changes
	--self:RegisterEvent("SPELLS_CHANGED", "TalentCheck")
    --self:RegisterEvent("LEARNED_SPELL_IN_TAB", "TalentCheck")
    self:RegisterEvent("PLAYER_TALENT_UPDATE", "TalentCheck")
    
    --Detect when player dies and reserect
    self:RegisterEvent("PLAYER_DEAD", "PlayerDead")
    
    --Detect when a player zone in and out, used to register for BAG_UPDATE events
    self:RegisterEvent("PLAYER_ENTERING_WORLD", "PlayerEnteringWorld")
    self:RegisterEvent("PLAYER_LEAVING_WORLD", "PlayerLeavingWorld")
    
    --Detect when trading with a vendor to auto-buy Ankhs
    self:RegisterEvent("MERCHANT_SHOW", "MerchantShow")

end


function BRAnkhTimer:OnDisable()
    self:DebugPrint("OnDisable")
    
    --Unregister all events
    self:UnregisterAllEvents()
    
    --Stop the timer if is running
    if (timerHandle ~= nil) then
		--Turn off the timer
		self:DebugPrint("Turn off the timer")
		self:CancelTimer(timerHandle)
		timerHandle = nil
    end
    
    --Update the data object
	self:RefreshText()
    
end


--###############################################################
--Event callbacks
--###############################################################


--Delayed OnEnable
--This delay is needed since its not possible to read any talent information at the OnEnable state, probably the UI is not fully initialized at that point
function BRAnkhTimer:DelayedOnEnable()
	self:DebugPrint("DelayedOnEnable")
	
	--Calculate start values
	ankhsInBags = self:NumberOfAnkhsInBags()
	reincarnationCooldown = self:getReincarnationCooldown()
	
	--Check if Glyph of Renewed Life is present
	glyphPresent = self:isGlyphPresent()
	
	--Check if we need to start the cooldown timer
	if ((time() - self.db.profile.reincarnationLastUsed) < reincarnationCooldown ) then
		self:DebugPrint("Still on cooldown, start timer")
		--Start timer to exectue the countdown
		timerHandle = self:ScheduleRepeatingTimer("CooldownTimerEvent", 1)
	elseif (self:isReincarnationReady() == false) then
		--Special case, the spell is on cooldown but not according to the time stored in self.db.profile.reincarnationLastUsed
		--This can happen the first time the addon is installed or if the spell is used without the addon present (addon disabled, another computer etc)
		self.db.profile.reincarnationLastUsed = 0
		--Start timer to exectue the countdown
		timerHandle = self:ScheduleRepeatingTimer("CooldownTimerEvent", 1)
	end
	
	--Update the data object
	self:RefreshText()
	
end


--Check if the cooldown of the Reincarnation spell has changed
--This is run after changing dual talent specs
function BRAnkhTimer:TalentCheck()
	self:DebugPrint("TalentCheck")
	
	--Check if Glyph of Renewed Life is present
	glyphPresent = self:isGlyphPresent()
	
	--Get the cooldown
	local cooldown = self:getReincarnationCooldown()
	
	--check if the cooldown has changed
	if (cooldown ~= reincarnationCooldown) then
		self:DebugPrint("Reincarnation cooldown time changed")
		reincarnationCooldown = cooldown
		
		--Update the data object
		self:RefreshText()
		
	end
	
end


--Player is alive, check if it is due to using the Reincarnation spell
function BRAnkhTimer:PlayerAlive()
	self:DebugPrint("PlayerAlive")
	
	--Stop listening to the PLAYER_ALIVE event
	self:UnregisterEvent("PLAYER_ALIVE")
	
	--Unfortenatly I can not check the start time using the Blizz function start, duration, enable = GetSpellCooldown(spellName, BOOKTYPE_SPELL) here
	--The PLAYER_ALIVE message probably happens before the new values are set for the spell
	--Start a timer event and check the values 1 seconds after the PLAYER_ALIVE message
	timerHandle = self:ScheduleTimer("PlayerAliveTimerEvent", 1, time())

	
end


--Called after PLAYER_ALIVE event with a short delay need to allow the game set spell values
function BRAnkhTimer:PlayerAliveTimerEvent(reserectionTime)
	--Check if player alive is due to using the Reincarnation spell
	if (self:isReincarnationReady() == false) then
		self:DebugPrint("PlayerAliveTimerEvent:Player used Ankh to reserect")
		
		--Store the time when Ankh was used
		self.db.profile.reincarnationLastUsed = reserectionTime
		
		--Start timer to exectue the countdown
		timerHandle = self:ScheduleRepeatingTimer("CooldownTimerEvent", 1)
		
		--Update the data object
		self:RefreshText()
		
	else
		self:DebugPrint("PlayerAliveTimerEvent:Ankh was not used")
	end

end


--Player is dead
function BRAnkhTimer:PlayerDead()
	self:DebugPrint("PlayerDead")
	
	--Check if reincarnation is available and ready to be used, then the next PLAYER_ALIVE event might by due to player using the reincarnation spell
	if (self:isReincarnationReady() == true) then
		self:RegisterEvent("PLAYER_ALIVE", "PlayerAlive")
	end
	
end


--Player entering world, register for BAG_UPDATE events
function BRAnkhTimer:PlayerEnteringWorld()
	self:DebugPrint("PlayerEnteringWorld")
	
	self:RegisterEvent("BAG_UPDATE", "BagUpdate")
end


--Player leaving world, unregister for BAG_UPDATE events to keep load down
function BRAnkhTimer:PlayerLeavingWorld()
	self:DebugPrint("PlayerLeavingWorld")
	
	self:UnregisterEvent("BAG_UPDATE")
end


--Bags are updated, check if the number of Ankhs has changed
function BRAnkhTimer:BagUpdate()
	self:DebugPrint("BagUpdate")
	
	local ankhs = self:NumberOfAnkhsInBags()
	--check if it differs for before the event
	if (ankhs ~= ankhsInBags) then
		self:DebugPrint("Number of Ankhs in the bag changed")
		ankhsInBags = ankhs
		
	--Update the data object
	self:RefreshText()
		
	end

end


--Trading with a vendor, try to auto-buy Ankhs
function BRAnkhTimer:MerchantShow()
	self:DebugPrint("MerchantShow")

	--Check if the current player is a shaman, if not never auto-buy anything	
	if select(2, UnitClass("player")) ~= "SHAMAN" then
		return
	end
	
	--Check if Glyph of Renewed Life is present
	--I do a new check here due to lack of a good event to hook this up to. This will update the state every now and again
	glyphPresent = self:isGlyphPresent()

	if (self.db.profile.autoBuyAnkhs == true and (not glyphPresent) ) then
	
		--Get the number of items the vendor has
		local numItems = GetMerchantNumItems()
		
		--Check each item
		for index = 1, numItems, 1 do
			-- Get the item link
			local itemLink = GetMerchantItemLink(index)
		
			if (itemLink ~= nil) then
				--Extract the itemstring
		    	local found, _, itemString = string.find(itemLink, "^|c%x+|H(.+)|h%[.*%]")
		    
		        --split up the item sting to find the itemID
		    	local _, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId = strsplit(":", itemString)
		    	
		    	--Check if its an Ankh
		    	if (itemId == "17030") then
		    		self:DebugPrint("Vendor is selling Ankhs")
		    		--Auto-buy if needed
		    		if (ankhsInBags < self.db.profile.maxAnkhs) then
		    			self:DebugPrint("By a few Ankhs from the vendor")
		    			BuyMerchantItem(index , (self.db.profile.maxAnkhs - ankhsInBags))
		    			
		    			--Print a message, its nice if the use know what addon is buying stuff
		    			DEFAULT_CHAT_FRAME:AddMessage(L["ankh_restocking_msg"])
		    			
		    		end
		    	end
		    	
			end
			
		end
		
	end	
	
end


--Called once per second when the reincarnation spell is on cooldown
function BRAnkhTimer:CooldownTimerEvent()
	--Check if the timer should be turned off
	if (self:isReincarnationReady() == true) then
		--Turn off the timer
		self:DebugPrint("Turn off the timer")
		self:CancelTimer(timerHandle)
		timerHandle = nil
		
		--Adjust any delta time left on the calculated cooldown, or the timer might stop at 00:01
		self.db.profile.reincarnationLastUsed = time() - reincarnationCooldown
		
	end
	
    --Update the data object
	self:RefreshText()

end



--###############################################################
--Data Broker methods
--###############################################################

--Update the data object with new text
function BRAnkhTimer:RefreshText()
	--self:DebugPrint("UpdateDataObject")
	
    --Update the data object
    self.dataObject.text = self:getDisplayString()
    
    --Inform listeners that they should refesh them self
    --self:Fire("Broker_AnkhTimer_Refresh")
   
end


--Create the text string to display
function BRAnkhTimer:getDisplayString()
    local s = ""
    
    --Create the count part of the string
    if (self.db.profile.showCount == true and (not glyphPresent) ) then
    	s = s .. format("|cff%s%d|r Ankh(s)", self:GetHexColor(ankhsInBags/10), ankhsInBags)
    	--self:DebugPrint("Data object ankh text = " .. s)
    end
    
    --Create the cooldown part of the string
    if (self.db.profile.showCooldown == true) then
    	local deltaTime = time() - self.db.profile.reincarnationLastUsed
    	--Check if the delta time is less the the reincarnation cooldown time
    	if (deltaTime < reincarnationCooldown) then
    		--Show current cooldown
    		local remainingTime = reincarnationCooldown - deltaTime
    		s = s .. format(" |cff%s%s|r", self:GetHexColor(1-remainingTime/reincarnationCooldown), date("%M:%S", remainingTime))
    		--self:DebugPrint("Data object ankh + cooldown text = " .. s)
    	elseif (self:isReincarnationReady() == false) then
    		--Special case, spell is on cooldown but not possible to calculate
    		--TODO: Make better presentation of this case
    		s = s .. format(" " .. L["unknown"], COLOR_HEX_WHITE)
    	end
    end
    
    return s
    
end   


--Called from LDB when the tooltip should be updated
function BRAnkhTimer.dataObject.OnTooltipShow(tooltip)

	local reincarnationLastUsed = BRAnkhTimer.db.profile.reincarnationLastUsed
	
	tooltip:AddLine("AnkhTimer")
 	--The date Reincarnation was used the last time
    if (reincarnationLastUsed ~= 0) then
  		tooltip:AddDoubleLine(L["last_used"], format("|cff%s%s|r", COLOR_HEX_WHITE, date("%c", reincarnationLastUsed)))
    else
    	tooltip:AddDoubleLine(L["last_used"], format(L["unknown"], COLOR_HEX_WHITE))
    end
    	 
    --The number of Ankhs in the bags
    tooltip:AddDoubleLine("Ankhs: ", format("|cff%s%s|r", COLOR_HEX_WHITE, ankhsInBags))

    --The number of minutes until Ankh can be used again
    local deltaTime = time() - reincarnationLastUsed
    if( deltaTime < reincarnationCooldown ) then
    	local remainingTime = reincarnationCooldown - deltaTime
        tooltip:AddDoubleLine(L["remaining"], format("|cff%s%s|r", COLOR_HEX_WHITE, date("%M:%S", remainingTime)))
    elseif ( (reincarnationLastUsed == 0) and (BRAnkhTimer:isReincarnationReady() == false) ) then
    	--Special case, Spell is on cooldown but not possible to caculate
    	tooltip:AddDoubleLine(L["remaining"], format(L["unknown"], COLOR_HEX_WHITE))
    else
        tooltip:AddDoubleLine(L["remaining"], format(L["ankh_is_ready"], COLOR_HEX_WHITE))
    end
    
    --The reincrnation spell cooldown
    tooltip:AddDoubleLine(L["ability_cooldown"], format("|cff%s%s|r", COLOR_HEX_WHITE, ((reincarnationCooldown/60).. L["min"])))
    
    tooltip:AddLine(format(L["right_click"], COLOR_HEX_YELLOW) .. L["open_options_menu"])
 
end


--Called when klicking on the addon
function BRAnkhTimer.dataObject.OnClick(self, button)
	--Open the configuraton menu if right button was used
	if (button == "RightButton") then
		InterfaceOptionsFrame_OpenToCategory(BRAnkhTimer.optionsFrame)
	end
end




--###############################################################
--Debug prints
--###############################################################

function BRAnkhTimer:DebugPrint(text)
	--check if debug is enabled
	if (text ~= nil and self.db.profile.debugPrints == true) then
		self:Print(text)
	end
end



--###############################################################
--Configuration callback methods
--###############################################################


function BRAnkhTimer:SetMaxAnkhs(info, value) 
	self:DebugPrint("SetMaxAnkhs")

	if (value ~=nil) then
		--Try and convert to a number
		num = tonumber(value)
		if (num ~=nil) then
			self.db.profile.maxAnkhs = num 
		else
			self:Print(L["must_be_number"])
		end
	end
 
end
            	

function BRAnkhTimer:GetMaxAnkhs(info) 
	self:DebugPrint("GetMaxAnkhs")
	local str = tostring(self.db.profile.maxAnkhs)
	if (str ~= nil) then
		return str
	else
		return ""
	end

end


function BRAnkhTimer:TestPrint()
	self:Print("Printing out internal state of the addon")
	
	self:Print("Local variables")
	self:Print("--ankhsInBags = " .. ankhsInBags)
	self:Print("--reincarnationCooldown = " .. reincarnationCooldown)
	self:Print("--Glyph = " .. self:BooleanToString(glyphPresent))
	if (timerHandle ~= nil) then
		self:Print("--timerHandle = " .. timerHandle)
	else
		self:Print("--timerHandle = nil")
	end
	
	self:Print("DB settings")
	self:Print("--debugPrints = " .. self:BooleanToString(self.db.profile.debugPrints))
	self:Print("--showCount = " .. self:BooleanToString(self.db.profile.showCount))
	self:Print("--showCooldown = " .. self:BooleanToString(self.db.profile.showCooldown))
	self:Print("--autoBuy = " .. self:BooleanToString(self.db.profile.autoBuyAnkhs))
	self:Print("--maxAnkhs = " .. self.db.profile.maxAnkhs)
	self:Print("--reincarnationLastUsed = " .. self.db.profile.reincarnationLastUsed)
	
	self:Print("Calculated values")
	self:Print("--ankhsInBags = " .. self:NumberOfAnkhsInBags())
	self:Print("--reincarnationCooldown = " .. self:getReincarnationCooldown())
	self:Print("--isReincarnationReady = " .. self:BooleanToString(self:isReincarnationReady()))
	self:Print("--Glyph = " .. self:BooleanToString(self:isGlyphPresent()))
	
end


--Simple help function for printing out booleans
function BRAnkhTimer:BooleanToString(value)
	if (value == nil) then 
		return "nil"
	elseif (value == true) then
		return "true"
	else
		return "false"
	end
end



--###############################################################
--Help functions
--###############################################################

--TODO make these fuctions local?????

--Count the number of Ankhs currently in the bags
function BRAnkhTimer:NumberOfAnkhsInBags()
	
	local ankhs = 0
	
	--Check all 5 bags
	for bag=0, 4, 1 do
		
		--Number of slots in the bag
        local bagSize = GetContainerNumSlots(bag)
        
        if (bagSize > 0) then
			--Check each slot in the bag
            for slot=1, bagSize, 1 do	
	            --Get the item link
                local itemLink = GetContainerItemLink(bag, slot)
           		
                --Ceck that the slot actually contains someting
            	if (itemLink ~=nil) then
            		
            	    --Extract the itemstring
                	local found, _, itemString = string.find(itemLink, "^|c%x+|H(.+)|h%[.*%]")
                
	                --split up the item sting to find the itemID
	            	local _, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId = strsplit(":", itemString)
	            	
	            	--Check if its an Ankh
	            	if (itemId == "17030") then
	            		local texture, itemCount = GetContainerItemInfo(bag, slot)
	            		ankhs  = ankhs + itemCount
	            	end	
            	end 
   			end
   		end
    
   	end
    
    return ankhs
    
end


--Get the Reincarnation cooldown based in the current talents set
function BRAnkhTimer:getReincarnationCooldown()

	--Find the number of talents set for Improved Self-Reincarnation
    local talentPoints = 0
    local tabNumber = 3
    
    --Check each tab
    repeat
 
    	--Check each talent on that tab
    	for talentIndex=1, GetNumTalents(tabNumber), 1 do
    	 
    		local name, icon, iconx, icony, currRank, maxRank = GetTalentInfo(tabNumber, talentIndex)
    		
    		if (name ~= nil) then
    			-- Is it the Improved Reincarnation talent?
                if (name == L["imp_reincarnation"]) then
                    talentPoints = currRank
                    break
                end
            end
    		
    	end
    
    	tabNumber = tabNumber -1
    until (talentPoints ~=0 or tabNumber < 1)
    
    --Convert to seconds    
    return (3600 - (10 * 60 * talentPoints))

end


--Check if the reincarnation spell is ready do be used
function BRAnkhTimer:isReincarnationReady()
	local spellName = GetSpellInfo(20608)
	local start, duration, enabled = GetSpellCooldown(spellName)
	if (start == 0) then
		return true
	else
		return false
	end
end


--Check if the "Glyph of Renewed Life" is present
function BRAnkhTimer:isGlyphPresent()
	
	--Check each Glyph socket
	for i=1, GetNumGlyphSockets(), 1 do
	
		local enabled, glyphType, glyphSpellID, icon = GetGlyphSocketInfo(i)
			
		if (glyphSpellID == 43385) then
			return true
		end
		
	end
	
	return false

end


--Return a color code base on the the percentage figure
function BRAnkhTimer:GetHexColor(p)
	if (p >= 1) then
		--Green
	 	return COLOR_HEX_GREEN
    elseif (p > 0.5) then
    	--Yellow
    	return COLOR_HEX_YELLOW
    elseif (p > 0.05) then
    	--Orange 
      return COLOR_HEX_ORANGE
    else
    	--Red
    	return COLOR_HEX_RED
    end
end

--TODO:
--Problems when switching between dual specs, all talent controls are made before the Ankh talent is updated
--Sometime the timer stops on the bar (text not refreshed when timer is stopped)

--Suggestions:
--Review the colors used (divide in small steps going from red to green)
--Dialog warning when low on Ankhs
--Use colors on the tooptip
--Dynamic tooltip to shown countdown of cooldown
--Notification when the coodown is up (what kind of notification?)

--Questions:
--Any troubles when another class run this addon by misstake?
--Test with other data broker displays
--Any issues when low-level shamans are using this addon?
--Some special cases during and after PLAYER_ENTERING_WORLD when the number of slots in the bags are set to 0, some cashing issues
--Are database values stored unique for each charcter?

--Limitations
--Not possible to calculate and show the cooldown if on cooldown when the addon is used first time
--If reincarnation is used with the addon not present (disabled, on another computer etc) its not possible to calculate and show the cooldown

--Knowbugs:
--

